@@ -9,7 +9,7 @@ namespace TextDocumentDocumentLink  {
   struct Params {
     // The document to provide document links for.
     lsTextDocumentIdentifier textDocument;
-	MAKE_SWAP_METHOD(Params, textDocument)
+        MAKE_SWAP_METHOD(Params, textDocument)
   };
 
 };
@@ -24,21 +24,21 @@ struct lsDocumentLink {
   // The range this link applies to.
   lsRange range;
   // The uri this link points to. If missing a resolve request is sent later.
-  boost::optional<lsDocumentUri> target;
+  optional<lsDocumentUri> target;
+
+  optional<lsp::Any> data;
 
-  boost::optional<lsp::Any> data;
-	
   MAKE_SWAP_METHOD(lsDocumentLink, range, target, data)
-	
+
 };
 MAKE_REFLECT_STRUCT(lsDocumentLink, range, target,data);
 
 
-DEFINE_REQUEST_RESPONSE_TYPE(td_links, TextDocumentDocumentLink::Params, lsDocumentLink, "textDocument/documentLink");
+DEFINE_REQUEST_RESPONSE_TYPE(td_links, TextDocumentDocumentLink::Params, std::vector<lsDocumentLink>, "textDocument/documentLink");
 
 
 /**
  * The document link resolve request is sent from the client to the server to resolve the target of a given document link.
  */
-DEFINE_REQUEST_RESPONSE_TYPE(td_linkResolve, std::vector<lsDocumentLink>, lsDocumentLink, "documentLink/resolve");
+DEFINE_REQUEST_RESPONSE_TYPE(td_linkResolve, lsDocumentLink, lsDocumentLink, "documentLink/resolve");
 
