@@ -17,16 +17,27 @@
 #include "config.h"
 #endif
 
+#if __cplusplus < 201703L
+
 #ifdef HAVE_LSP
 #include <boost/optional.hpp>
 #include <boost/none.hpp>
 using boost::optional;
+#define nullopt boost::none
 using boost::make_optional;
 #else
 #include "optional.hpp"
 #define boost nonstd
 using nonstd::optional;
 using nonstd::nullopt;
+using nonstd::make_optional;
+#endif
+
+#else
+#include <optional>
+using std::optional;
+using std::nullopt;
+using std::make_optional;
 #endif
 
 using std::make_pair;
