@@ -56,9 +56,8 @@ inline void openpipeout()
   int fd=intcast(settings::getSetting<Int>("outpipe"));
   if(!pipeout && fd >= 0) pipeout=fdopen(fd,"w");
   if(!pipeout) {
-    ostringstream buf;
-    buf << "Cannot open outpipe " << fd;
-    reportError(buf);
+    cerr << "Cannot open outpipe " << fd << endl;
+    exit(-1);
   }
 }
 
@@ -417,7 +416,7 @@ public:
 
   void Read(bool &val) {string t; readwhite(t); val=(t == "true");}
   void Read(Int& val) {*stream >> val;}
-  void Read(double& val) {*stream >> val;}
+  void Read(double& val);
   void Read(pair& val) {*stream >> val;}
   void Read(triple& val) {*stream >> val;}
   void Read(char& val) {stream->get(val);}
