@@ -50,6 +50,7 @@
 #define GC_THREADS
 #ifdef __clang__
 #define GC_ATTR_EXPLICIT
+#define GC_NOEXCEPT
 #endif
 #include <gc.h>
 
@@ -133,7 +134,11 @@ GC_CONTAINER(vector);
 GC_CONTAINER(deque);
 
 template <typename T, typename Container = vector<T> >
-struct stack : public std::stack<T, Container>, public gc {
+struct stack : public std::stack<T, Container>, public gc {};
+
+template <typename T, typename S >
+struct pair : public std::pair<T, S>, public gc {
+  pair(T t, S s) : std::pair<T,S>(t,s) {}
 };
 
 #define PAIR_ALLOC gc_allocator<std::pair<CONST Key,T> > /* space */
