@@ -35,7 +35,7 @@
 # include <sys/stat.h>
 #endif
 
-#if defined(CONSOLE_LOG) && defined(MSWIN32) && defined(_MSC_VER)
+#if defined(CONSOLE_LOG) && defined(MSWIN32) && !defined(__GNUC__)
 # include <io.h>
 #endif
 
@@ -248,7 +248,7 @@ STATIC void GC_init_size_map(void)
         GC_size_map[i] = ROUNDED_UP_GRANULES(i);
 #       ifndef _MSC_VER
           GC_ASSERT(GC_size_map[i] < TINY_FREELISTS);
-          /* Seems to tickle bug in VC++ 2008 for AMD64 */
+          /* Seems to tickle bug in VC++ 2008 for x64 */
 #       endif
     }
     /* We leave the rest of the array to be filled in on demand. */
@@ -730,7 +730,8 @@ GC_API void GC_CALL GC_get_heap_usage_safe(GC_word *pheap_size,
   {
     char *p;
     char *end_of_content;
-    unsigned namelen;
+    size_t namelen;
+
 #   ifndef NO_GETENV
       p = getenv(name); /* try the standard getenv() first */
       if (p != NULL)
@@ -781,12 +782,18 @@ GC_API int GC_CALL GC_is_init_called(void)
   {
     if (GC_find_leak && !skip_gc_atexit) {
 #     ifdef THREADS
-        GC_in_thread_creation = TRUE; /* OK to collect from unknown thread. */
-        GC_gcollect();
-        GC_in_thread_creation = FALSE;
-#     else
-        GC_gcollect();
+        /* Check that the thread executing at-exit functions is     */
+        /* the same as the one performed the GC initialization,     */
+        /* otherwise the latter thread might already be dead but    */
+        /* still registered and this, as a consequence, might       */
+        /* cause a signal delivery fail when suspending the threads */
+        /* on platforms that do not guarantee ESRCH returned if     */
+        /* the signal is not delivered.                             */
+        /* It should also prevent "Collecting from unknown thread"  */
+        /* abort in GC_push_all_stacks().                           */
+        if (!GC_is_main_thread() || !GC_thread_is_registered()) return;
 #     endif
+      GC_gcollect();
     }
   }
 #endif
@@ -1199,7 +1206,8 @@ GC_API void GC_CALL GC_init(void)
 #   if ALIGNMENT > GC_DS_TAGS
       /* Adjust normal object descriptor for extra allocation.  */
       if (EXTRA_BYTES != 0)
-        GC_obj_kinds[NORMAL].ok_descriptor = (word)(-ALIGNMENT) | GC_DS_LENGTH;
+        GC_obj_kinds[NORMAL].ok_descriptor =
+                        ((~(word)ALIGNMENT) + 1) | GC_DS_LENGTH;
 #   endif
     GC_exclude_static_roots_inner(beginGC_arrays, endGC_arrays);
     GC_exclude_static_roots_inner(beginGC_obj_kinds, endGC_obj_kinds);
@@ -1213,12 +1221,6 @@ GC_API void GC_CALL GC_init(void)
         /* entirety as part of the root set.  This will grow them to    */
         /* maximum size, and is generally not desirable.                */
 #   endif
-#   if defined(SEARCH_FOR_DATA_START)
-        GC_init_linux_data_start();
-#   endif
-#   if defined(NETBSD) && defined(__ELF__)
-        GC_init_netbsd_elf();
-#   endif
 #   if !defined(THREADS) || defined(GC_PTHREADS) \
         || defined(NN_PLATFORM_CTR) || defined(NINTENDO_SWITCH) \
         || defined(GC_WIN32_THREADS) || defined(GC_SOLARIS_THREADS)
@@ -1255,6 +1257,21 @@ GC_API void GC_CALL GC_init(void)
       GC_ASSERT(!((word)GC_stackbottom HOTTER_THAN (word)GC_approx_sp()));
 #   endif
     GC_init_headers();
+#   if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED) \
+       && (defined(SEARCH_FOR_DATA_START) || defined(NETBSD))
+      LOCK(); /* just to set GC_lock_holder */
+#   endif
+#   ifdef SEARCH_FOR_DATA_START
+      /* For MPROTECT_VDB, the temporary fault handler should be        */
+      /* installed first, before the write fault one in GC_dirty_init.  */
+      if (GC_REGISTER_MAIN_STATIC_DATA()) GC_init_linux_data_start();
+#   elif defined(NETBSD) && defined(__ELF__)
+      if (GC_REGISTER_MAIN_STATIC_DATA()) GC_init_netbsd_elf();
+#   endif
+#   if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED) \
+       && (defined(SEARCH_FOR_DATA_START) || defined(NETBSD))
+      UNLOCK();
+#   endif
 #   ifndef GC_DISABLE_INCREMENTAL
       if (GC_incremental || 0 != GETENV("GC_ENABLE_INCREMENTAL")) {
 #       if defined(BASE_ATOMIC_OPS_EMULATED) || defined(CHECKSUMS) \
@@ -1279,6 +1296,7 @@ GC_API void GC_CALL GC_init(void)
     /* Add initial guess of root sets.  Do this first, since sbrk(0)    */
     /* might be used.                                                   */
       if (GC_REGISTER_MAIN_STATIC_DATA()) GC_register_data_segments();
+
     GC_bl_init();
     GC_mark_init();
     {
@@ -1340,16 +1358,6 @@ GC_API void GC_CALL GC_init(void)
 #       else
           GC_thr_init();
 #       endif
-#       ifdef PARALLEL_MARK
-          /* Actually start helper threads.     */
-#         if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED)
-            UNLOCK();
-#         endif
-          GC_start_mark_threads_inner();
-#         if defined(GC_ASSERTIONS) && defined(GC_ALWAYS_MULTITHREADED)
-            LOCK();
-#         endif
-#       endif
 #   endif
     COND_DUMP;
     /* Get black list set up and/or incremental GC started */
@@ -1377,9 +1385,8 @@ GC_API void GC_CALL GC_init(void)
 #   if defined(PARALLEL_MARK) || defined(THREAD_LOCAL_ALLOC) \
        || (defined(GC_ALWAYS_MULTITHREADED) && defined(GC_WIN32_THREADS) \
            && !defined(GC_NO_THREADS_DISCOVERY))
-        /* Make sure marker threads are started and thread local */
-        /* allocation is initialized, in case we didn't get      */
-        /* called from GC_init_parallel.                         */
+        /* Make sure thread local allocation is initialized, in */
+        /* case we did not get called from GC_init_parallel().  */
         GC_init_parallel();
 #   endif /* PARALLEL_MARK || THREAD_LOCAL_ALLOC */
 
@@ -1402,7 +1409,7 @@ GC_API void GC_CALL GC_enable_incremental(void)
       LOCK();
       if (!GC_incremental) {
         GC_setpagesize();
-        /* if (GC_no_win32_dlls) goto out; Should be win32S test? */
+        /* TODO: Should we skip enabling incremental if win32s? */
         maybe_install_looping_handler(); /* Before write fault handler! */
         if (!GC_is_initialized) {
           UNLOCK();
@@ -1445,23 +1452,22 @@ GC_API void GC_CALL GC_enable_incremental(void)
   GC_init();
 }
 
-#if defined(THREADS)
-  GC_API void GC_CALL GC_start_mark_threads(void)
-  {
-#   if defined(PARALLEL_MARK) && defined(CAN_HANDLE_FORK) \
-       && !defined(THREAD_SANITIZER)
-      /* TSan does not support threads creation in the child process.   */
+GC_API void GC_CALL GC_start_mark_threads(void)
+{
+#   ifdef PARALLEL_MARK
       IF_CANCEL(int cancel_state;)
+      DCL_LOCK_STATE;
 
       DISABLE_CANCEL(cancel_state);
+      LOCK();
       GC_start_mark_threads_inner();
+      UNLOCK();
       RESTORE_CANCEL(cancel_state);
 #   else
       /* No action since parallel markers are disabled (or no POSIX fork). */
       GC_ASSERT(I_DONT_HOLD_LOCK());
 #   endif
-  }
-#endif
+}
 
   GC_API void GC_CALL GC_deinit(void)
   {
@@ -1472,7 +1478,9 @@ GC_API void GC_CALL GC_enable_incremental(void)
 #       if !defined(CONSOLE_LOG) || defined(MSWINCE)
           DeleteCriticalSection(&GC_write_cs);
 #       endif
-        DeleteCriticalSection(&GC_allocate_ml);
+#       ifndef GC_PTHREADS
+          DeleteCriticalSection(&GC_allocate_ml);
+#       endif
 #     endif
     }
   }
@@ -2146,13 +2154,9 @@ GC_API void * GC_CALL GC_call_with_alloc_lock(GC_fn_type fn, void *client_data)
     void * result;
     DCL_LOCK_STATE;
 
-#   ifdef THREADS
-      LOCK();
-#   endif
+    LOCK();
     result = (*fn)(client_data);
-#   ifdef THREADS
-      UNLOCK();
-#   endif
+    UNLOCK();
     return(result);
 }
 
@@ -2166,6 +2170,8 @@ GC_API void * GC_CALL GC_call_with_stack_base(GC_stack_base_func fn, void *arg)
       base.reg_base = (void *)GC_save_regs_in_stack();
       /* Unnecessarily flushes register stack,          */
       /* but that probably doesn't hurt.                */
+#   elif defined(E2K)
+      base.reg_base = NULL; /* not used by GC currently */
 #   endif
     result = fn(&base, arg);
     /* Strongly discourage the compiler from treating the above */
@@ -2245,9 +2251,11 @@ STATIC void GC_do_blocking_inner(ptr_t data, void * context GC_ATTR_UNUSED)
         GC_blocked_sp = GC_save_regs_in_stack();
 #   else
         GC_blocked_sp = (ptr_t) &d; /* save approx. sp */
-#   endif
-#   ifdef IA64
-        GC_blocked_register_sp = GC_save_regs_in_stack();
+#       ifdef IA64
+            GC_blocked_register_sp = GC_save_regs_in_stack();
+#       elif defined(E2K)
+            (void)GC_save_regs_in_stack();
+#       endif
 #   endif
 
     d -> client_data = (d -> fn)(d -> client_data);
@@ -2284,6 +2292,8 @@ STATIC void GC_do_blocking_inner(ptr_t data, void * context GC_ATTR_UNUSED)
     sb -> mem_base = GC_stackbottom;
 #   ifdef IA64
       sb -> reg_base = GC_register_stackbottom;
+#   elif defined(E2K)
+      sb -> reg_base = NULL;
 #   endif
     return &GC_stackbottom; /* gc_thread_handle */
   }
@@ -2369,7 +2379,6 @@ GC_API GC_word GC_CALL GC_get_gc_no(void)
 #ifdef THREADS
   GC_API int GC_CALL GC_get_parallel(void)
   {
-    /* GC_parallel is initialized at start-up.  */
     return GC_parallel;
   }
 
@@ -2405,6 +2414,37 @@ GC_API GC_word GC_CALL GC_get_gc_no(void)
     UNLOCK();
     return fn;
   }
+
+# ifdef STACKPTR_CORRECTOR_AVAILABLE
+    GC_INNER GC_sp_corrector_proc GC_sp_corrector = 0;
+# endif
+
+  GC_API void GC_CALL GC_set_sp_corrector(
+                                GC_sp_corrector_proc fn GC_ATTR_UNUSED)
+  {
+#   ifdef STACKPTR_CORRECTOR_AVAILABLE
+      DCL_LOCK_STATE;
+
+      LOCK();
+      GC_sp_corrector = fn;
+      UNLOCK();
+#   endif
+  }
+
+  GC_API GC_sp_corrector_proc GC_CALL GC_get_sp_corrector(void)
+  {
+#   ifdef STACKPTR_CORRECTOR_AVAILABLE
+      GC_sp_corrector_proc fn;
+      DCL_LOCK_STATE;
+
+      LOCK();
+      fn = GC_sp_corrector;
+      UNLOCK();
+      return fn;
+#   else
+      return 0; /* unsupported */
+#   endif
+  }
 #endif /* THREADS */
 
 /* Setter and getter functions for the public R/W function variables.   */
@@ -2413,8 +2453,9 @@ GC_API GC_word GC_CALL GC_get_gc_no(void)
 
 GC_API void GC_CALL GC_set_oom_fn(GC_oom_func fn)
 {
-    GC_ASSERT(NONNULL_ARG_NOT_NULL(fn));
     DCL_LOCK_STATE;
+
+    GC_ASSERT(NONNULL_ARG_NOT_NULL(fn));
     LOCK();
     GC_oom_fn = fn;
     UNLOCK();
