@@ -1248,6 +1248,8 @@ void initSettings() {
                             "Emulate unimplemented SVG shading", true));
   addOption(new boolSetting("prc", 0,
                             "Embed 3D PRC graphics in PDF output", false));
+  addOption(new boolSetting("v3d", 0,
+                            "Embed 3D V3D graphics in PDF output", false));
   addOption(new boolSetting("toolbar", 0,
                             "Show 3D toolbar in PDF output", true));
   addOption(new boolSetting("axes3", 0,
@@ -1387,7 +1389,7 @@ void initSettings() {
                                       "Allow read from other directory",
                                       &globalRead, true));
   addSecureSetting(new stringSetting("outname", 'o', "name",
-                                     "Alternative output directory/filename"));
+                                     "Alternative output directory/file prefix"));
   addOption(new stringOption("cd", 0, "directory", "Set current directory",
                              &startpath));
 
@@ -1407,10 +1409,13 @@ void initSettings() {
                             "Input code over multiple lines at the prompt"));
   addOption(new boolSetting("xasy", 0,
                             "Interactive mode for xasy"));
+#ifdef HAVE_LSP
   addOption(new boolSetting("lsp", 0, "Interactive mode for the Language Server Protocol"));
-  addOption(new boolSetting("wsl", 0, "Run asy under the Windows Subsystem for Linux."));
   addOption(new envSetting("lspport", ""));
   addOption(new envSetting("lsphost", "127.0.0.1"));
+#endif
+
+  addOption(new boolSetting("wsl", 0, "Run asy under the Windows Subsystem for Linux"));
 
   addOption(new boolSetting("wait", 0,
                             "Wait for child processes to finish before exiting"));
@@ -1470,8 +1475,6 @@ void initSettings() {
                             60.0));
   addOption(new realSetting("framerate", 0, "frames/s", "Animation speed",
                             30.0));
-  addOption(new realSetting("framedelay", 0, "ms",
-                            "Additional frame delay", 0.0));
   addOption(new realSetting("resizestep", 0, "step", "Resize step", 1.2));
   addOption(new IntSetting("digits", 0, "n",
                            "Default output file precision", 7));
